# Generated by Django 2.2.13 on 2021-10-04 13:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0019_location_image'),
    ]

    operations = [
        migrations.AddField(
            model_name='location',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.2.3 on 2021-03-18 11:28

from django.db import migrations

def load_sources(apps, schema_editor):
    SourceType = apps.get_model('measurements', 'SourceType')
    SourceType.objects.create(code="cmems_wms")

class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0012_auto_20210318_1121'),
    ]

    operations = [
        migrations.RunPython(load_sources)
    ]


# Generated by Django 2.2.13 on 2021-03-23 16:22

from django.db import migrations, models

sql_create_view = """
create or replace view measurements_grafana_view as
 SELECT mm."timestamp",
    mm.value,
    ml.label AS location,
    ml.id AS location_id,
    mst.code AS station_code,
    mst.label AS station,
    mst.id AS station_id,
    n.id AS network_id,
    n.code AS network_code,
    n.label AS network_label,
    st_x(ml.geo) AS longitude,
    st_y(ml.geo) AS latitude,
    mp.code AS parameter,
    mp.id AS parameter_id,
    ms.height,
    ms.id AS serie_id
   FROM measurements_measure mm
     LEFT JOIN measurements_serie ms ON mm.serie_id = ms.id
     LEFT JOIN measurements_station mst ON ms.station_id = mst.id
     LEFT JOIN measurements_network n ON mst.network_id = n.id
     LEFT JOIN measurements_location ml ON ms.location_id = ml.id
     LEFT JOIN measurements_parameter mp ON ms.parameter_id = mp.id;
"""

sql_drop_view = """drop view if exists measurements_grafana_view;"""

sql_grant = """grant select on measurements_grafana_view to public;"""


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0016_parametermapping_source_parameter_uom'),
    ]

    operations = [
        migrations.AddField(
            model_name='station',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.RunSQL(sql_drop_view),
        migrations.RunSQL(sql_create_view),
        migrations.RunSQL(sql_grant),
    ]



# Generated by Django 2.2.13 on 2021-03-19 08:49

from django.db import migrations, models
import measurements.models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0015_auto_20210319_0846'),
    ]

    operations = [
        migrations.AddField(
            model_name='parametermapping',
            name='source_parameter_uom',
            field=models.CharField(blank=True, help_text='unit of measure based on python-pint unit registry (eg. meter, kelvin)', max_length=25, null=True, validators=[measurements.models.validate_uom]),
        ),
    ]



# Generated by Django 2.2.3 on 2021-01-14 08:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0007_auto_20210113_1138'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='measure',
            unique_together={('serie', 'timestamp')},
        ),
    ]



# Generated by Django 2.2.13 on 2021-04-27 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0017_station_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='station',
            name='image',
            field=models.ImageField(blank=True, null=True, upload_to=''),
        ),
    ]



from django.test import RequestFactory, TestCase, Client
from django.contrib.auth.models import AnonymousUser, User

from measurements.writeapi import write
from measurements.models import Measure

LP_INVALID = """mymeas,mytag1=1 value=21 1463689680000000000"""

LP_VALID = """waves,parameter=AirPress_SL,sensor=WVP2,station=PTF value=21 1463689680000000000
waves,parameter=AirPress_SL,sensor=WVP2,station=PTF value=22 1463689690000000000
waves,parameter=AirPress_SL,sensor=WVP2,station=PTF value=23 1463689700000000000
"""

class WriteAPITest(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    def test_invalid_serie(self):
        request = self.factory.post('/measurements/api',
                                    data=LP_INVALID,
                                    content_type='application/octet-stream'
                                    )
        request.user = AnonymousUser()
        self.assertRaises(ValueError, write, request)

    def test_write(self):
        request = self.factory.post('/measurements/api',
                                    data=LP_VALID,
                                    content_type='application/octet-stream'
                                    )
        # TODO:
        request.user = AnonymousUser()
        write(request)
        self.assertTrue(Measure.objects.exists())

"""
create table measurements_duplicates as 
select id, m.serie_id, m.timestamp
from measurements_measure m, 
    (select serie_id, timestamp from measurements_measure 
            group by (serie_id, timestamp) having count(id) >1) as d 
where m.serie_id=d.serie_id and m.timestamp=d.timestamp;

-- test result --
select
mm.id as measure_id,
mm.serie_id,
ml.label AS location,
mst.source_id,
mst.network_id,
mp.code AS parameter,
mm."timestamp",
mm.value
from measurements_measure mm
LEFT JOIN measurements_serie ms ON mm.serie_id = ms.id
LEFT JOIN measurements_station mst ON ms.station_id = mst.id
LEFT JOIN measurements_location ml ON mst.location_id = ml.id
LEFT JOIN measurements_parameter mp ON ms.parameter_id = mp.id
where mm.id in (select id from measurements_duplicates) -- where date_part('year', timestamp)>=2020 )
order by mm.serie_id, timestamp, mm.id;

begin;
delete from measurements_measure
where id in 
(
select 
distinct on (mm.serie_id, timestamp) 
mm.id
from measurements_measure mm
where mm.id in (select id from measurements_duplicates)
order by mm.serie_id, timestamp, mm.id
);

commit;

"""


from django.core.management.base import BaseCommand

from measurements.sources.cmems_wms import CMEMSWmsDataSource
from measurements.models import SourceType
from measurements.utils import get_serie, load_serie
from measurements import  ureg, Q_

class Command(BaseCommand):
    help = "Command to import CMEMS WMS data."

    def add_arguments(self, parser):
        parser.add_argument(
            '--days',
            type=int,
            default=30
        )
        parser.add_argument(
            '--daysgap',
            type=int,
            default=0
        )

    def handle(self, *args, **options):
        days = options['days']
        daysgap = options['daysgap']

        ps = SourceType.objects.get(code='cmems_wms')
        for station in ps.station_set.filter(status='active'):
            self.stdout.write("Loading DataSource/Station {} ... ".format(station))
            apiclient = CMEMSWmsDataSource(station.uri)

            self.stdout.write("\tLooking for registered layers/parameters ...")
            for pm in station.parametermapping_set.all():
                parameter = pm.parameter
                source_uom = pm.source_parameter_uom
                uom = parameter.uom
                print(source_uom, uom, parameter)
                layer = pm.source_parameter_label
                self.stdout.write("\tLooking for registered series ...")
                for series in station.serie_set.filter(parameter=parameter).all():
                    location = series.location
                    # print(series)
                    self.stdout.write(
                        "\tLoading series {}. Layer {}. Location {} ... ".format(series.id, layer, location),
                        ending=''
                    )
                    point = [location.geo.x, location.geo.y]

                    df = apiclient.get_df(layer, point, timedelta=days, timegap=daysgap)
                    # convert uom if needed
                    if uom is not None and source_uom is not None and uom != source_uom:
                        df['value'] = Q_(df.value.array, source_uom).to(uom).magnitude
                    if df.shape[0] > 0:
                        load_serie(df['value'].copy(), series.id)
                        self.stdout.write("nrecords {}".format(df.shape[0]), ending='')
                        self.stdout.write("[OK]")
                    else:
                        self.stdout.write("[FAILED]")



# Source classes should not include reference to django
# This will facilitate use of functionalities within other packages/projects
# Django-related tasks (eg. uom conversion, fieldname mapping) should be moved
# into utilis or management commands

class BaseSource(object):
    def get_data(self):
        pass




# Generated by Django 2.2.13 on 2021-08-02 14:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0018_station_image'),
    ]

    operations = [
        migrations.AddField(
            model_name='location',
            name='image',
            field=models.ImageField(blank=True, null=True, upload_to=''),
        ),
    ]



# Generated by Django 2.2.3 on 2021-03-18 11:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0011_auto_20210318_1012'),
    ]

    operations = [
        migrations.CreateModel(
            name='ParameterMapping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('source_parameter_label', models.CharField(max_length=64)),
                ('data_source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='measurements.Station')),
                ('parameter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='measurements.Parameter')),
            ],
        ),
        migrations.AddField(
            model_name='station',
            name='parameters_mapping',
            field=models.ManyToManyField(through='measurements.ParameterMapping', to='measurements.Parameter'),
        ),
    ]



# import pandas as pd
from django.contrib.gis.db import models
from django.core.exceptions import ValidationError
from django.contrib.postgres.fields import ArrayField
from django.utils.text import slugify
from django_pandas.io import read_frame
from psqlextra.manager import PostgresManager
from measurements import ureg, Q_
from pint.errors import UndefinedUnitError
import numpy as np


def validate_uom(value):
    if value is None:
        return value
    try:
        return ureg(value).units
    except UndefinedUnitError:
        pass
    raise ValidationError("The provided unit of measure ({}) is not identifiable. Please provide a valid name.".format(value))


class Parameter(models.Model):
    code = models.CharField(max_length=100)
    uri = models.URLField(blank=True, null=True, help_text="Reference to a controlled vocabulary (eg. NERC)")
    label = models.CharField(max_length=150, blank=True, null=True)
    uom = models.CharField(max_length=25,
                           validators=[validate_uom],
                           help_text="unit of measure based on python-pint unit registry (eg. meter, kelvin)",
                           blank=True, null=True)

    objects = models.Manager()
    extra = PostgresManager()

    def __str__(self):
        if self.uom is None:
            return u'{}'.format(self.code)
        else:
            return u'{} ({})'.format(self.code, self.uom)


class Sensor(models.Model):
    code = models.CharField(max_length=100)
    label = models.CharField(max_length=150, blank=True, null=True)

    objects = models.Manager()
    extra = PostgresManager()

    def __str__(self):
        return u'{}'.format(self.label)


class LocationManager(models.Manager):
    def get_flatjson(self):
        data = []
        qs = super().get_queryset()
        for l in qs:
            r = {'key': slugify(l.label),
                 'latitute': l.geo.centroid.y,
                 'longitude': l.geo.centroid.x,
                 'name': l.label
                 }
            data.append(r)
        return data


class Location(models.Model):
    label = models.CharField(max_length=150)
    description = models.TextField(null=True, blank=True)
    geo = models.PointField(srid=4326, null=True)
    elevation = models.IntegerField(null=True, blank=True)
    image = models.ImageField(null=True, blank=True)

    objects = LocationManager()
    extra = PostgresManager()

    def __str__(self):
        return u'{}'.format(self.label)


class Network(models.Model):
    code = models.CharField(max_length=100)
    label = models.CharField(max_length=150, blank=True, null=True)

    def __str__(self):
        if self.label is not None:
            return u'{}'.format(self.label)
        else:
            return u'{}'.format(self.code)


class SourceType(models.Model):
    code = models.CharField(max_length=100, unique=True)

    def __str__(self):
        return u'{}'.format(self.code)


class Station(models.Model):
    code = models.CharField(max_length=100)
    label = models.CharField(max_length=150, null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    image = models.ImageField(null=True, blank=True)
    network = models.ForeignKey(Network, on_delete=models.CASCADE, null=True, blank=True)
    # location has been moved directly within Series model
    # this location field should provide current Station location
    location = models.ForeignKey(Location, on_delete=models.CASCADE, null=True, blank=True)
    source = models.ForeignKey(SourceType, on_delete=models.CASCADE, null=True, blank=True)
    status = models.CharField(max_length=255,
                              choices=[(i, i) for i in ['active', 'decomissioned', 'in maintenance']], default='active')
    parameters_mapping = models.ManyToManyField(Parameter, through='ParameterMapping')
    # this allows to override SourceType endpoint
    uri = models.URLField(blank=True, null=True, help_text="DataSource/Station endpoint")

    objects = models.Manager()
    extra = PostgresManager()

    def __str__(self):
        l = getattr(self, 'label') or getattr(self, 'code')
        n = getattr(self, 'network', '')
        return u'{} {}'.format(l, n)

    class Meta:
        verbose_name = "DataSource/Station"
        verbose_name_plural = "DataSources/Stations"


class ParameterMapping(models.Model):
    data_source = models.ForeignKey(Station, on_delete=models.CASCADE)
    source_parameter_label = models.CharField(max_length=64,
                                              help_text="Parameter label or layer name published by the data provider")
    source_parameter_uom = models.CharField(max_length=25,
                                            validators=[validate_uom],
                                            help_text="unit of measure based on python-pint unit registry (eg. meter, kelvin)",
                                            blank=True, null=True)
    parameter = models.ForeignKey(Parameter, on_delete=models.CASCADE)


class Serie(models.Model):
    station = models.ForeignKey(Station, on_delete=models.CASCADE)
    sensor = models.ForeignKey(Sensor, on_delete=models.CASCADE)
    parameter = models.ForeignKey(Parameter, on_delete=models.CASCADE)
    height = models.FloatField(blank=True, null=True)
    location = models.ForeignKey(Location, on_delete=models.CASCADE, null=True, blank=True)
    # add below additional fields
    stats_mean = models.FloatField(blank=True, null=True)
    stats_outliers = ArrayField(models.IntegerField(), blank=True, null=True)

    objects = models.Manager()
    extra = PostgresManager()

    def set_mean(self, threshold=5):
        df = read_frame(Measure.objects.filter(serie=self),
                        index_col='timestamp')

        df['pandas'] = df['value'].rolling(window=5, center=True).median().fillna(method='bfill').fillna(method='ffill')

        difference = np.abs(df['value'] - df['pandas'])
        outlier_idx = difference > threshold

        # print df.loc[outlier_idx, 'value'].shape
        self.stats_mean = df.loc[~outlier_idx, 'value'].mean()
        self.save()
        return self.stats_mean

    def __str__(self):
        return u'{} - {} - {}'.format(self.station, self.parameter, self.sensor.label)


class Measure(models.Model):
    id = models.BigAutoField(primary_key=True)
    serie = models.ForeignKey(Serie, on_delete=models.CASCADE)
    timestamp = models.DateTimeField(db_index=True)
    value = models.FloatField()

    objects = models.Manager()
    extra = PostgresManager()

    class Meta:
        unique_together = ('serie', 'timestamp') #, 'value')



from setuptools import setup, find_packages

setup(
    name='django-measurements',
    version='0.29',
    # packages=['measurements'],
    packages=find_packages(),
    url='',
    license='GPL3',
    author='Stefano Menegon',
    author_email='stefano.menegon@cnr.it',
    description='',
    install_requires=['requests',
                      'pycryptodome',
                      'django-postgres-extra',
                      'django_pandas',
                      'geojson',
                      'colorbrewer',
                      'django-basicauth',
                      'lxml',
                      'beautifulsoup4',
                      'pint',
                      'django-leaflet']
)



# Generated by Django 2.2.3 on 2021-02-11 11:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0008_auto_20210114_0832'),
    ]

    operations = [
        migrations.AddField(
            model_name='location',
            name='elevation',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.2.13 on 2021-03-19 08:05

from django.db import migrations, models
from django.db.models import F

def update_null(apps, schema_editor):
    Parameter = apps.get_model('measurements', 'Parameter')
    Parameter.objects.filter(code__isnull=True).update(code=F('label'))

class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0013_auto_20210318_1128'),
    ]

    operations = [
        migrations.RunPython(update_null),
        migrations.AddField(
            model_name='parameter',
            name='unit',
            field=models.CharField(blank=True, help_text='unit of measure based on python-pint unit registry (eg. meter, kelvin)', max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name='parameter',
            name='code',
            field=models.CharField(max_length=100),
        ),
        migrations.AlterField(
            model_name='parameter',
            name='label',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AlterField(
            model_name='parameter',
            name='uri',
            field=models.URLField(blank=True, help_text='Reference to a controlled vocabulary (eg. NERC)', null=True),
        ),
        migrations.AlterField(
            model_name='parametermapping',
            name='source_parameter_label',
            field=models.CharField(help_text='Parameter label or layer name published by the data provider', max_length=64),
        ),
    ]



from django.conf import settings
# from django.contrib import admin
from django.contrib.gis import admin
from .models import Parameter, Sensor, Location, Serie, Measure, Network, SourceType, Station, ParameterMapping
from leaflet.admin import LeafletGeoAdmin

load_google = False
try:
    settings.MAP_WIDGETS['GOOGLE_MAP_API_KEY'] 
    from mapwidgets.widgets import GooglePointFieldInlineWidget
    from django.contrib.gis.db.models import GeometryField
    formfield_overrides = {
        GeometryField: {"widget": GooglePointFieldInlineWidget}
    }


    class LocationAdmin(admin.ModelAdmin):
        formfield_overrides = formfield_overrides

except (KeyError, AttributeError) as e:
    formfield_overrides = {}


    class LocationAdmin(LeafletGeoAdmin):
        display_raw = True


class ParameterAdmin(admin.ModelAdmin):
    pass


class SensorAdmin(admin.ModelAdmin):
    pass


class NetworkAdmin(admin.ModelAdmin):
    pass


class SerieAdmin(admin.ModelAdmin):
    list_display = ('station', 'location', 'sensor', 'parameter', 'height')
    list_filter = ('station', 'sensor', 'parameter')


class ParameterMappingInline(admin.TabularInline):
    model = ParameterMapping


class StationAdmin(admin.ModelAdmin):
    inlines = [ParameterMappingInline]
    pass


class SourceTypeAdmin(admin.ModelAdmin):
    pass


class MeasureAdmin(admin.ModelAdmin):
    list_display = ('serie', 'timestamp', 'value')


admin.site.register(Parameter, ParameterAdmin)
admin.site.register(Sensor, SensorAdmin)
admin.site.register(Location, LocationAdmin)
admin.site.register(Network, NetworkAdmin)
admin.site.register(Serie, SerieAdmin)
admin.site.register(Station, StationAdmin)
admin.site.register(SourceType, SourceTypeAdmin)
admin.site.register(Measure, MeasureAdmin)



# Generated by Django 2.2.3 on 2021-01-13 11:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0006_station_status'),
    ]

    operations = [
        migrations.AlterField(
            model_name='measure',
            name='id',
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 2.2.3 on 2021-03-18 10:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0010_serie_location'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='station',
            options={'verbose_name': 'DataSource/Station', 'verbose_name_plural': 'DataSources/Stations'},
        ),
        migrations.AddField(
            model_name='station',
            name='uri',
            field=models.URLField(blank=True, help_text='DataSource/Station endpoint', null=True),
        ),
    ]



# Generated by Django 2.2.3 on 2021-03-17 16:35

from django.db import migrations, models
import django.db.models.deletion

sql_create_view = """
create or replace view measurements_grafana_view as
 SELECT mm."timestamp",
    mm.value,
    ml.label AS location,
    ml.id AS location_id,
    mst.label AS station,
    mst.id AS station_id,
    st_x(ml.geo) AS longitude,
    st_y(ml.geo) AS latitude,
    mp.code AS parameter,
    mp.id AS parameter_id,
    ms.height,
    ms.id AS serie_id
   FROM measurements_measure mm
     LEFT JOIN measurements_serie ms ON mm.serie_id = ms.id
     LEFT JOIN measurements_station mst ON ms.station_id = mst.id
     LEFT JOIN measurements_location ml ON ms.location_id = ml.id
     LEFT JOIN measurements_parameter mp ON ms.parameter_id = mp.id;
"""

sql_drop_view = """drop view if exists measurements_grafana_view;"""

sql_grant = """grant select on measurements_grafana_view to public;"""

sql_update = """
    update measurements_serie
        set location_id=mst.location_id
        from measurements_station mst
        where measurements_serie.station_id = mst.id;
"""


class Migration(migrations.Migration):

    dependencies = [
        ('measurements', '0009_location_elevation'),
    ]

    operations = [
        migrations.RunSQL(sql_drop_view),
        migrations.AddField(
            model_name='serie',
            name='location',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='measurements.Location'),
        ),
        migrations.RunSQL(sql_update),
        migrations.RunSQL(sql_create_view),
        migrations.RunSQL(sql_grant),
    ]



import logging
from datetime import date
import pandas as pd
from measurements.sources.base import BaseSource
from owslib.wms import WebMapService
from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)


def response2df(response):
    res = BeautifulSoup(response.read(), 'xml')
    data = []
    for r in res.FeatureInfoResponse.findAll('FeatureInfo'):
        v = r.value.text
        t = r.time.text
        data.append([t, v])

    df = pd.DataFrame(data, columns=["time", 'value'])
    df.replace({'none': None}, inplace=True)
    df['value'] = df.value.astype(float)
    df['time'] = pd.to_datetime(df.time)
    df.set_index('time', inplace=True)
    return df


def get_timeindex(timepositions):
    dr = None
    for _t in timepositions:
        t = _t.strip()
        if '/' in t:
            start, end, freq = t.split('/')
            # print(pd.to_datetime(start))
            _dr = pd.date_range(start.strip(), end.strip(), freq='1D')
            # print(start, end, period)
        # positions.appen
        # print(t)
        else:
            try:
                _dr = pd.date_range(t, periods=1, freq='1D')
            except ValueError as err:
                logger.warning("Skip invalid time position: {0}".format(err))
                continue
        if dr is None:
            dr = _dr
        else:
            dr = dr.union(_dr)
    return dr


def align_time(time, time_index, tolerance='30d'):
    id = time_index.get_loc(time,
                            method='nearest',
                            tolerance=tolerance)
    return time_index[id]

# TODO: create view  SELECT DISTINCT l.label AS __text,
#     l.id AS __value
#    FROM measurements_station s
#      LEFT JOIN measurements_location l ON s.location_id = l.id
#   WHERE s.status::text = 'active'::text
#   ORDER BY l.label;

class CMEMSWmsDataSource(BaseSource):
    def __init__(self, url, version='1.3.0'):
        self.url = url

        self.wms = WebMapService(url,
                                 version=version)
        self.response = None
        self.df = None
        self._time_indexes = {}

    def get_timeindex(self, layer):
        if layer not in self._time_indexes:
            i = get_timeindex(self.wms[layer].timepositions)
            self._time_indexes[layer] = i
        return self._time_indexes.get(layer)

    def get_df(self, layer, point, reftime=None, elevation=None,
               timedelta=None, timegap=None):
        """
        Query the WMS data source using the GetFeatureInfo method. It supports Elevation and Time dimensions.

        :param layer: WMS layer
        :param point: lat/long 2D array
        :param reftime: WMS Time range (eg. '2020-01-01/2020-12-01') or single date when used in together with timedelta or timegap parameters (eg. 2020-12-01)
        :param elevation: specify an elevation/depth level. If None default elevation will be returned.
        :param timedelta: time delta in days
        :param timegap: time gap in days
        :return: DataFrame with a datetime index and a "value" column
        """
        if reftime is not None:
            _time = reftime
        else:
            _time = date.today().isoformat()
        if timedelta is not None:
            _end = pd.to_datetime(_time)
            if timegap is not None:
                _end = _end - pd.Timedelta(days=timegap)
            _start = _end - pd.Timedelta(days=timedelta)
            time_index = self.get_timeindex(layer)
            _start = align_time(_start, time_index)
            _end = align_time(_end, time_index)
            _time = "{}/{}".format(_start.isoformat()[:-6],
                                   _end.isoformat()[:-6])

        bbox = (point[0] - 0.1,
                point[1] - 0.1,
                point[0] + 0.1,
                point[1] + 0.1,
                )
        _options = {
            'layers': [layer],
            'srs': 'EPSG:4326',
            'bbox': bbox,
            'size': (3, 3),
            'format': 'image/jpeg',
            'query_layers': [layer],
            'info_format': "text/xml",
            'xy': (1, 1)
        }
        if _time is not None:
            _options['time'] = _time
        if elevation is not None:
            _options['elevation'] = elevation

        logger.debug("request {} {} {} {}".format(layer, point, _time, elevation))
        self.response = self.wms.getfeatureinfo(**_options)
        self.df = response2df(self.response)
        return self.df
